/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLChannel.h"
#include <nsCOMPtr.h>
#include <nsString.h>
#include <nsContentCID.h>
#include <nsIStreamListener.h>
#include <nsIWindowWatcher.h>
#include <nsIServiceManagerUtils.h>
#include <necko/nsNetUtil.h>

NS_IMPL_ISUPPORTS4(mozCellMLChannel, nsIRequest, nsIChannel, nsIDocumentObserver, mozCellMLIImportConsumer)
NS_IMPL_ISUPPORTS1(mozCellMLImportRequest, nsIDocumentObserver)

mozCellMLChannel::mozCellMLChannel(nsIChannel* aChan)
  : mChan(aChan)
{
}

mozCellMLChannel::~mozCellMLChannel()
{
  if (mDoc != nsnull)
  {
    mDoc->RemoveObserver(NS_STATIC_CAST(nsIDocumentObserver*, this));
    mDoc = nsnull;
  }

}

NS_IMETHODIMP
mozCellMLChannel::GetName(nsACString& name)
{
  return mChan->GetName(name);
}

NS_IMETHODIMP
mozCellMLChannel::IsPending(PRBool* pend)
{
  return mChan->IsPending(pend);
}

NS_IMETHODIMP
mozCellMLChannel::GetStatus(nsresult* stat)
{
  return mChan->GetStatus(stat);
}

NS_IMETHODIMP
mozCellMLChannel::Cancel(nsresult ret)
{
  return mChan->Cancel(ret);
}

NS_IMETHODIMP
mozCellMLChannel::Suspend()
{
  return mChan->Suspend();
}

NS_IMETHODIMP
mozCellMLChannel::Resume()
{
  return mChan->Resume();
}

NS_IMETHODIMP
mozCellMLChannel::SetLoadGroup(nsILoadGroup* aLoadGroup)
{
  NS_ASSERTION(doc == nsnull, "SetLoadGroup called after AsyncOpen()");
  mLoadGroup = aLoadGroup;
  return mChan->SetLoadGroup(aLoadGroup);
}

NS_IMETHODIMP
mozCellMLChannel::GetLoadGroup(nsILoadGroup** aLoadGroup)
{
  return mChan->GetLoadGroup(aLoadGroup);
}

NS_IMETHODIMP
mozCellMLChannel::SetLoadFlags(nsLoadFlags loadFlags)
{
  return mChan->SetLoadFlags(loadFlags);
}

NS_IMETHODIMP
mozCellMLChannel::GetLoadFlags(nsLoadFlags* loadFlags)
{
  return mChan->GetLoadFlags(loadFlags);
}

NS_IMETHODIMP
mozCellMLChannel::GetOriginalURI(nsIURI** aURI)
{
  if (mURI == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Failing because we can't GetOriginalURI!\n");
#endif
    return NS_ERROR_FAILURE;
  }
  *aURI = mURI;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLChannel::SetOriginalURI(nsIURI* aURI)
{
  mURI = aURI;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLChannel::GetURI(nsIURI** aURI)
{
  return mChan->GetURI(aURI);
}

NS_IMETHODIMP
mozCellMLChannel::GetOwner(nsISupports** owner)
{
  return mChan->GetOwner(owner);
}

NS_IMETHODIMP
mozCellMLChannel::SetOwner(nsISupports* owner)
{
  return mChan->SetOwner(owner);
}

NS_IMETHODIMP
mozCellMLChannel::SetNotificationCallbacks(nsIInterfaceRequestor* notcb)
{
  return mChan->SetNotificationCallbacks(notcb);
}

NS_IMETHODIMP
mozCellMLChannel::GetNotificationCallbacks(nsIInterfaceRequestor** notcb)
{
  return mChan->GetNotificationCallbacks(notcb);
}

NS_IMETHODIMP
mozCellMLChannel::GetSecurityInfo(nsISupports** secInfo)
{
  return mChan->GetSecurityInfo(secInfo);
}

NS_IMETHODIMP
mozCellMLChannel::GetContentType(nsACString& ctype)
{
  return mChan->GetContentType(ctype);
}

NS_IMETHODIMP
mozCellMLChannel::SetContentType(const nsACString& ctype)
{
  return mChan->SetContentType(ctype);
}

NS_IMETHODIMP
mozCellMLChannel::GetContentCharset(nsACString& ccharset)
{
  return mChan->GetContentCharset(ccharset);
}

NS_IMETHODIMP
mozCellMLChannel::SetContentCharset(const nsACString& ccharset)
{
  return mChan->SetContentCharset(ccharset);
}

NS_IMETHODIMP
mozCellMLChannel::GetContentLength(PRInt32* length)
{
  return mChan->GetContentLength(length);
}

NS_IMETHODIMP
mozCellMLChannel::SetContentLength(PRInt32 length)
{
  return mChan->SetContentLength(length);
}

NS_IMETHODIMP
mozCellMLChannel::Open(nsIInputStream** iis)
{
#ifdef DEBUG_MODE
  printf("Failing because we can't Open()...\n");
#endif
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
mozCellMLChannel::AsyncOpen(
                            nsIStreamListener* aListener,
                            nsISupports* context
                           )
{
  // We ignore listener, as we don't provide data back to the consumer, instead
  // it goes to the document, which eventually gets displayed in a chrome
  // window or similar.

#ifdef DEBUG_MODE
  printf("mozCellMLChannel AsyncOpen()\n");
#endif
  // This is a hack to make the 1.7.x tree work...

  nsresult rv;

  static NS_DEFINE_CID(kXMLDocumentCID,
                       NS_XMLDOCUMENT_CID
                      );

  mDoc = do_CreateInstance("@mozilla.org/xml/xml-document;1");
#ifdef DEBUG_MODE
  printf("Creating a document...\n");
#endif
  if (mDoc == nsnull)
  {
    mDoc = do_CreateInstance(kXMLDocumentCID);
  }
  if (mDoc == nsnull)
  {
#ifdef DEBUG_MODE
    printf("mDoc could not be created!!!\n");
#endif
    return NS_ERROR_OUT_OF_MEMORY;
  }
  nsCOMPtr<nsIStreamListener> aDocListener;

#ifdef DEBUG_MODE
  printf("Starting document load for underlying document.\n");
#endif
  rv = mDoc->StartDocumentLoad(
                               "loadAsData",
                               mChan,
                               mLoadGroup,
                               nsnull,
                               getter_AddRefs(aDocListener),
                               PR_FALSE,
                               nsnull
                              );
  NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG_MODE
  printf("Calling AsyncOpen()...\n");
#endif
  rv = mChan->AsyncOpen(aDocListener, context);
  NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG_MODE
  printf("Adding the observer...\n");
#endif

  // We hold a reference to ourselves while loading...
  AddRef();

  mDoc->AddObserver(NS_STATIC_CAST(nsIDocumentObserver*, this));
  return NS_OK;
}

NS_IMPL_NSIDOCUMENTOBSERVER_CORE_STUB(mozCellMLChannel)
NS_IMPL_NSIDOCUMENTOBSERVER_REFLOW_STUB(mozCellMLChannel)
NS_IMPL_NSIDOCUMENTOBSERVER_STATE_STUB(mozCellMLChannel)
NS_IMPL_NSIDOCUMENTOBSERVER_CONTENT(mozCellMLChannel)
NS_IMPL_NSIDOCUMENTOBSERVER_STYLE_STUB(mozCellMLChannel)

void
mozCellMLChannel::BeginLoad(nsIDocument* aDocument)
{
}

void
mozCellMLChannel::EndLoad(nsIDocument* aDocument)
{
#ifdef DEBUG_MODE
  printf("EndLoad called...\n");
#endif
  nsresult rv;
  nsCOMPtr<nsIDOMDocument> dd = do_QueryInterface(aDocument);
  model = new mozCellML();

  rv = model->SetDom(dd);
  if (rv != NS_OK)
    return;
#ifdef DEBUG_MODE
  printf("Fetching imports from the model...\n");
#endif
  rv = model->FetchImports(getter_AddRefs(mImportRequests));
  if (rv != NS_OK)
    return;
#ifdef DEBUG_MODE
  printf("Counting imports...\n");
#endif
  mImportRequests->GetLength(&nRemaining);
#ifdef DEBUG_MODE
  printf("There are %u imports remaining to load...\n", nRemaining);
#endif

  if (nRemaining == 0)
  {
    model->Initialise();
    nsCOMPtr<nsIWindowWatcher> ww = do_GetService(NS_WINDOWWATCHER_CONTRACTID);
    nsCOMPtr<nsIDOMWindow> window;
    ww->OpenWindow(NULL, "chrome://mozcellml/content/modelview.xul",
                   NULL, "chrome,dialog=no", model, getter_AddRefs(window));
  }

  nsCOMPtr<nsISimpleEnumerator> importE;
  mImportRequests->Enumerate(getter_AddRefs(importE));
  PRBool moreEls;
  while (importE->HasMoreElements(&moreEls), moreEls)
  {
    nsCOMPtr<nsISupports> importSupport;
    importE->GetNext(getter_AddRefs(importSupport));
    nsCOMPtr<mozCellMLImportRequest>
      importRequest(do_QueryInterface(importSupport));
    importRequest->StartLoad(mLoadGroup, this);
  }

  Release();
}

void
mozCellMLChannel::ImportLoaded(nsIDOMElement* import,
                               nsIDOMDocument* imported)
{
#ifdef DEBUG_MODE
  printf("mozCellMLChannel: Import loaded.\n");
#endif
  nRemaining--;
  model->ImportLoaded(import, imported);
  if (nRemaining == 0)
  {
    model->Initialise();
    nsCOMPtr<nsIWindowWatcher> ww = do_GetService(NS_WINDOWWATCHER_CONTRACTID);
    nsCOMPtr<nsIDOMWindow> window;
    ww->OpenWindow(NULL, "chrome://mozcellml/content/modelview.xul",
                   NULL, "chrome,dialog=no", model, getter_AddRefs(window));
  }
}

mozCellMLImportRequest::mozCellMLImportRequest(nsIDOMElement* iimportNode)
  : importNode(iimportNode)
{
}

mozCellMLImportRequest::~mozCellMLImportRequest()
{
  if (mDoc != nsnull)
  {
    mDoc->RemoveObserver(NS_STATIC_CAST(nsIDocumentObserver*, this));
    mDoc = nsnull;
  }
}

nsresult
mozCellMLImportRequest::StartLoad(
                                  nsILoadGroup* aLoadGroup,
                                  mozCellMLChannel* isuperChannel
                                 )
{
  nsresult rv;

  // Get the URL we are supposed to load...
  nsAutoString importURL;
#ifdef DEBUG_MODE
  printf("Fetching load URL...\n");
#endif
  importNode->GetAttributeNS(NS_LITERAL_STRING("http://www.w3.org/1999/xlink"),
                             NS_LITERAL_STRING("href"),
                             importURL);

  nsCAutoString cimportURL;
  CopyUTF16toUTF8(importURL, cimportURL);
#ifdef DEBUG_MODE
  printf("Load URL = %s\n", cimportURL.get());
#endif
  // Make a URL object...
  nsCOMPtr<nsIURI> impURI = do_CreateInstance(NS_STANDARDURL_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = impURI->SetSpec(cimportURL);
  NS_ENSURE_SUCCESS(rv, rv);
  
  // Create a channel to load the URL...
  nsCOMPtr<nsIIOService> pService(do_GetIOService(&rv));
#ifdef DEBUG_MODE
  printf("Making new channel to load import...\n");
#endif
  rv = pService->NewChannelFromURI(impURI, getter_AddRefs(mChannel));

  superChannel = isuperChannel;
  static NS_DEFINE_CID(
                       kXMLDocumentCID,
                       NS_XMLDOCUMENT_CID
                      );
  mDoc = do_CreateInstance("@mozilla.org/xml/xml-document;1");
  if (mDoc == nsnull)
    mDoc = do_CreateInstance(kXMLDocumentCID);
  if (mDoc == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Failing because we can't make an XML Document.\n");
#endif
    return NS_ERROR_OUT_OF_MEMORY;
  }

  nsCOMPtr<nsIStreamListener> aDocListener;

#ifdef DEBUG_MODE
  printf("Starting import document load...\n");
#endif
  rv = mDoc->StartDocumentLoad(
                               "loadAsData",
                               mChannel,
                               aLoadGroup,
                               nsnull,
                               getter_AddRefs(aDocListener),
                               PR_FALSE,
                               nsnull
                              );
  NS_ENSURE_SUCCESS(rv, rv);
#ifdef DEBUG_MODE
  printf("Calling AsyncOpen() for import...\n");
#endif
  rv = mChannel->AsyncOpen(aDocListener, nsnull);
  NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG_MODE
  printf("Adding observer...\n");
#endif
  mDoc->AddObserver(NS_STATIC_CAST(nsIDocumentObserver*, this));
  return NS_OK;
}

NS_IMPL_NSIDOCUMENTOBSERVER_CORE_STUB(mozCellMLImportRequest)
NS_IMPL_NSIDOCUMENTOBSERVER_REFLOW_STUB(mozCellMLImportRequest)
NS_IMPL_NSIDOCUMENTOBSERVER_STATE_STUB(mozCellMLImportRequest)
NS_IMPL_NSIDOCUMENTOBSERVER_CONTENT(mozCellMLImportRequest)
NS_IMPL_NSIDOCUMENTOBSERVER_STYLE_STUB(mozCellMLImportRequest)

void
mozCellMLImportRequest::BeginLoad(nsIDocument* aDocument)
{
}

void
mozCellMLImportRequest::EndLoad(nsIDocument* aDocument)
{
#ifdef DEBUG_MODE
  printf("Import load complete.\n");
#endif
  nsCOMPtr<nsIDOMDocument> dd = do_QueryInterface(aDocument);
  superChannel->ImportLoaded(importNode, dd);

  //  Break cycle...
  superChannel = nsnull;
}
