/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLVariable.h"
#include <prdtoa.h>
#include <nsPrintfCString.h>
#include <nsIArray.h>
#include <nsArray.h>
#include <nsInterfaceHashtable.h>
#include <nsPrintfCString.h>

NS_IMPL_ISUPPORTS1(mozCellMLVariable, mozCellMLIVariable)

mozCellMLVariable::mozCellMLVariable(nsIDOMElement* var,
                                     mozCellML* mcm)
  : seen(PR_FALSE), done(PR_FALSE), degree(1),
    variable(var), mozcellml(mcm)
{
}

#define COMPLAIN(model,args) \
{\
  nsPrintfCString complainCString args; \
  nsAutoString complainString; \
  CopyUTF8toUTF16(complainCString, complainString); \
  return model->ValidationError(complainString); \
}

void
mozCellMLVariable::Initialise(nsInterfaceHashtable<nsStringHashKey,
                              mozCellMLVariable>& allVariables,
                              nsIArray* components)
{
  nsresult ret = DetermineType(allVariables, components, mozcellml);

  // Break the cycle...
  mozcellml = nsnull;

  if (ret != NS_OK)
  {
    type = 123;
    return;
  }
}

mozCellMLVariable::~mozCellMLVariable()
{
}

nsresult
FilterNodeListByType(nsIDOMNodeList* nodelist, nsIArray** oarray,
                     PRUint16 keepType)
{
  nsIMutableArray *array;
  NS_NewArray(&array);

  PRUint32 len;
  nodelist->GetLength(&len);
  for (PRUint32 i = 0; i < len; i++)
  {
    nsCOMPtr<nsIDOMNode> dn;
    nodelist->Item(i, getter_AddRefs(dn));
    PRUint16 nodeType;
    if (dn->GetNodeType(&nodeType) != NS_OK)
      continue;
    if (nodeType != keepType)
      continue;
    array->AppendElement(dn, PR_FALSE);
  }

  // All ownership is handled by the caller...
  *oarray = array;

  return NS_OK;
}

nsresult
FilterNodeListByTypeNS(nsIDOMNodeList* nodelist, nsIArray** oarray,
                       PRUint16 keepType,
                       const nsAString& ns)
{
  nsIMutableArray *array;
  NS_NewArray(&array);

  PRUint32 len;
  nodelist->GetLength(&len);
  for (PRUint32 i = 0; i < len; i++)
  {
    nsCOMPtr<nsIDOMNode> dn;
    nodelist->Item(i, getter_AddRefs(dn));
    PRUint16 nodeType;
    if (dn->GetNodeType(&nodeType) != NS_OK)
      continue;
    if (nodeType != keepType)
      continue;

    nsAutoString xname;
    dn->GetNamespaceURI(xname);
    if (!xname.Equals(ns))
      continue;

    array->AppendElement(dn, PR_FALSE);
  }

  // All ownership is handled by the caller...
  *oarray = array;

  return NS_OK;
}

nsresult
FilterNodeListByNameAndTypeNS(
                              nsIDOMNodeList* nodelist,
                              nsIArray** oarray,
                              PRUint16 keepType,
                              const nsAString& ns,
                              const nsAString& nname
                             )
{
  nsIMutableArray *array;
  NS_NewArray(&array);

  PRUint32 len;
  nodelist->GetLength(&len);
  for (PRUint32 i = 0; i < len; i++)
  {
    nsCOMPtr<nsIDOMNode> dn;
    nodelist->Item(i, getter_AddRefs(dn));
    PRUint16 nodeType;
    if (dn->GetNodeType(&nodeType) != NS_OK)
      continue;
    if (nodeType != keepType)
      continue;

    nsAutoString xname, nname2;
    nname2 = nname;
    dn->GetLocalName(xname);
    if (!xname.Equals(nname2))
      continue;
    dn->GetNamespaceURI(xname);
    if (!xname.Equals(ns))
      continue;

    array->AppendElement(dn, PR_FALSE);
  }

  // All ownership is handled by the caller...
  *oarray = array;

  return NS_OK;
}

nsresult
FilterNodeListByNameAndTypeNS2(
                              nsIDOMNodeList* nodelist,
                              nsIArray** oarray,
                              PRUint16 keepType,
                              const nsAString& ns1,
                              const nsAString& ns2,
                              const nsAString& nname
                             )
{
  nsIMutableArray *array;
  NS_NewArray(&array);

  PRUint32 len;
  nodelist->GetLength(&len);
  for (PRUint32 i = 0; i < len; i++)
  {
    nsCOMPtr<nsIDOMNode> dn;
    nodelist->Item(i, getter_AddRefs(dn));
    PRUint16 nodeType;
    if (dn->GetNodeType(&nodeType) != NS_OK)
      continue;
    if (nodeType != keepType)
      continue;

    nsAutoString xname, nname2;
    nname2 = nname;
    dn->GetLocalName(xname);
    if (!xname.Equals(nname2))
      continue;
    dn->GetNamespaceURI(xname);
    if (!xname.Equals(ns1) &&
        !xname.Equals(ns2))
      continue;

    array->AppendElement(dn, PR_FALSE);
  }

  // All ownership is handled by the caller...
  *oarray = array;

  return NS_OK;
}

nsresult
mozCellMLVariable::Equals(mozCellMLIVariable* compareTo,
                          PRBool* result)
{
  nsAutoString myName, theirName;
  compareTo->GetName(theirName);
  GetName(myName);
  *result = (myName.Equals(theirName));
  return NS_OK;
}

nsresult
mozCellMLVariable::SetSeen(PRBool seen)
{
  this->seen = seen;
  return NS_OK;
}

nsresult
mozCellMLVariable::GetSeen(PRBool* seen)
{
  *seen = this->seen;
  return NS_OK;
}

nsresult
mozCellMLVariable::SetDone(PRBool done)
{
  this->done = done;
  return NS_OK;
}

nsresult
mozCellMLVariable::GetDone(PRBool* done)
{
  *done = this->done;
  return NS_OK;
}


nsresult
mozCellMLVariable::PrecomputeDependencies(
                                          mozCellML* model,
                                          nsInterfaceHashtable<nsStringHashKey,
                                          mozCellMLVariable>& allVariables
                                         )
{
  NS_NewArray(getter_AddRefs(dependencies));
  if (type == mozCellMLIVariable::CONSTANT)
    return NS_OK;
  // Now we just look on our right side to see what we depend on...
  return SearchForDependencies(model, allVariables, assignComponent,
                               assignFrom);
}

nsresult
mozCellMLVariable::OrderDependencies(
                                     mozCellML* mcm,
                                     nsCOMArray<mozCellMLVariable>& order
                                    )
{
  SetSeen(PR_TRUE);
  nsAutoString name;

  nsCOMPtr<nsISimpleEnumerator> deps;
  dependencies->Enumerate(getter_AddRefs(deps));
  PRUint32 i, l;
  for (
       i = 0, dependencies->GetLength(&l);
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> varSupports;
    deps->GetNext(getter_AddRefs(varSupports));
    nsCOMPtr<mozCellMLVariable> var(do_QueryInterface(varSupports));

    if (var->type != mozCellMLIVariable::SIMPLE)
      continue;

    PRBool wasSeen, wasDone;
    var->GetDone(&wasDone);
    if (wasDone)
      continue;
    var->GetSeen(&wasSeen);
    /*
     * This is the nasty case where the model defines something like a=2b,
     * b=2c, c=a or similar. In this case, we just fail.
     */
    if (wasSeen)
    {
      nsAutoString name1, name2;
      var->GetName(name1);
      GetName(name2);
      mcm->ValidationError(NS_LITERAL_STRING(
                               "Circular variable dependency created by ") +
                           name2 + NS_LITERAL_STRING(" => ") + name1);
      return NS_ERROR_FAILURE;
    }
    nsresult res = var->OrderDependencies(mcm, order);
    if (res != NS_OK)
      return res;
  }
  // All our children are done, so we are done...
  order.AppendObject(this);
  SetDone(PR_TRUE);
  return NS_OK;
}

nsresult
mozCellMLVariable::SearchForDependencies(
                                         mozCellML* model,
                                         nsInterfaceHashtable<nsStringHashKey,
                                         mozCellMLVariable>& allVariables,
                                         nsAString& component,
                                         nsIDOMElement* el)
{
  PRUint32 i, l;
  nsCOMPtr<nsIDOMNode> subnode;
  PRUint16 type;
  nsAutoString name;
  el->GetLocalName(name);
  if (name.Equals(NS_LITERAL_STRING("ci")) &&
      GetLHSVariableName(model, el, name, PR_FALSE) == NS_OK)
  {
    // We have a variable...
    nsCOMPtr<mozCellMLVariable> v;
    allVariables.Get(component + NS_LITERAL_STRING("::") + name,
                     getter_AddRefs(v));
    if (v == nsnull)
    {
      return model->ValidationError(
                  NS_LITERAL_STRING("Variable ") + name +
                  NS_LITERAL_STRING(" is used but not defined in component ") +
                  component
                                   );
    }
    if (v == this)
      return NS_OK;
    dependencies->AppendElement(v, PR_FALSE);
    return NS_OK;
  }

  nsCOMPtr<nsIDOMNodeList> children;
  el->GetChildNodes(getter_AddRefs(children));
  for (children->GetLength(&l), i = 0;
       i < l;
       i++)
  {
    children->Item(i, getter_AddRefs(subnode));
    subnode->GetNodeType(&type);
    if (type != nsIDOMNode::ELEMENT_NODE)
      continue;
    nsCOMPtr<nsIDOMElement> subelement = do_QueryInterface(subnode);
    nsresult ret = SearchForDependencies(model, allVariables, component,
                                         subelement);
    if (ret != NS_OK)
      return ret;
  }
  return NS_OK;
}

nsresult
mozCellMLVariable::ClassifyEqualsArgument(nsIDOMNode* node,
                                          EqualsArgumentTypes* xtype,
                                          PRUint32* xdegree,
                                          mozCellML* mcm)
{
  nsAutoString name;
  node->GetLocalName(name);

  if (name.Equals(NS_LITERAL_STRING("ci")))
    *xtype = CI;
  else if (name.Equals(NS_LITERAL_STRING("cn")))
    *xtype = CN;
  else if (name.Equals(NS_LITERAL_STRING("apply")))
  {
    // See whether this is a rate of change of a variable...
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    node->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 numSubnodes;
    subnodes->GetLength(&numSubnodes);
    if (numSubnodes == 0)
      return mcm->ValidationError(
                      NS_LITERAL_STRING(
                           "Found a math <apply> with no elements.")
                                 );
    nsCOMPtr<nsISupports> applyOpSupports;
    nsCOMPtr<nsISimpleEnumerator> subnodeEnum;
    subnodes->Enumerate(getter_AddRefs(subnodeEnum));
    subnodeEnum->GetNext(getter_AddRefs(applyOpSupports));
    nsCOMPtr<nsIDOMNode> applyOp(do_QueryInterface(applyOpSupports));
    applyOp->GetLocalName(name);
    if (name.Equals(NS_LITERAL_STRING("diff")) ||
        name.Equals(NS_LITERAL_STRING("partialdiff")))
    {
      *xtype = APPLY_DIFF;
      PRUint32 i;
      *xdegree = 1;
      for (i = 1; i < numSubnodes; i++)
      {
        nsCOMPtr<nsISupports> degSupports;
        subnodeEnum->GetNext(getter_AddRefs(degSupports));
        nsCOMPtr<nsIDOMElement> degEl(do_QueryInterface(degSupports));
        if (GetDegree(degEl, xdegree) == NS_OK)
          break;
      }
    }
    else
      *xtype = APPLY_MATH;
  }
  else if (name.Equals(NS_LITERAL_STRING("piecewise")))
  {
    *xtype = APPLY_MATH;
  }
  else
  {
    mcm->ValidationError(
            NS_LITERAL_STRING(
                  "Math apply nodes with an <eq /> operator must have ci, cn, or apply subnodes, not "
                             ) + name
                        );
  }
  return NS_OK;
}

nsresult
mozCellMLVariable::GetCNValue(nsIDOMNode* cn, double* value)
{
  nsCOMPtr<nsIDOMNodeList> subNodes;
  cn->GetChildNodes(getter_AddRefs(subNodes));
  nsCOMPtr<nsIArray> subText;
  FilterNodeListByType(subNodes, getter_AddRefs(subText),
                         nsIDOMNode::TEXT_NODE);
  PRUint32 numText;
  subText->GetLength(&numText);
  if (numText == 0)
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsISimpleEnumerator> subTextE;
  subText->Enumerate(getter_AddRefs(subTextE));
  nsCOMPtr<nsISupports> txtNodeSupports;
  subTextE->GetNext(getter_AddRefs(txtNodeSupports));
  nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
  nsAutoString valueStr;
  nsCAutoString valueCStr;
  txt->GetData(valueStr);
  valueStr.Trim(" \t\r\n\f");
  CopyUTF16toUTF8(valueStr, valueCStr);
  *value = PR_strtod(valueCStr.get(), nsnull);
  return NS_OK;
}

nsresult
mozCellMLVariable::GetDegree(
                             nsIDOMElement* possibleDeg,
                             PRUint32* degr
                            )
{
  nsAutoString localName;
  possibleDeg->GetLocalName(localName);
  if (!localName.Equals(NS_LITERAL_STRING("degree")))
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsIDOMNodeList> subNodes;
  possibleDeg->GetChildNodes(getter_AddRefs(subNodes));
  nsCOMPtr<nsIArray> subText;
  FilterNodeListByType(subNodes, getter_AddRefs(subText),
                         nsIDOMNode::TEXT_NODE);
  PRUint32 numText;
  subText->GetLength(&numText);
  if (numText == 0)
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsISimpleEnumerator> subTextE;
  subText->Enumerate(getter_AddRefs(subTextE));
  nsCOMPtr<nsISupports> txtNodeSupports;
  subTextE->GetNext(getter_AddRefs(txtNodeSupports));
  nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
  nsAutoString valueStr;
  nsCAutoString valueCStr;
  txt->GetData(valueStr);
  valueStr.Trim(" \t\r\n\f");
  CopyUTF16toUTF8(valueStr, valueCStr);
  *degr = strtoul(valueCStr.get(), nsnull, 10);
  return NS_OK;
}

nsresult
mozCellMLVariable::GetLHSVariableName(
                                      mozCellML* model,
                                      nsIDOMNode* left,
                                      nsAutoString& leftName,
                                      PRBool allowDiff)
{
  nsAutoString leftType;
  left->GetLocalName(leftType);
  if (leftType.Equals(NS_LITERAL_STRING("ci")))
  {
    nsCOMPtr<nsIDOMNodeList> subNodes;
    left->GetChildNodes(getter_AddRefs(subNodes));
    nsCOMPtr<nsIArray> subText;
    FilterNodeListByType(subNodes, getter_AddRefs(subText),
                         nsIDOMNode::TEXT_NODE);
    PRUint32 numText;
    subText->GetLength(&numText);
    if (numText == 0)
      return model->ValidationError(NS_LITERAL_STRING("No text nodes inside <ci>"));
    nsCOMPtr<nsISimpleEnumerator> subTextE;
    subText->Enumerate(getter_AddRefs(subTextE));
    nsCOMPtr<nsISupports> txtNodeSupports;
    subTextE->GetNext(getter_AddRefs(txtNodeSupports));
    nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
    txt->GetData(leftName);
    leftName.Trim(" \t\r\n\f");
  }
  else if (leftType.Equals(NS_LITERAL_STRING("apply")) &&
           allowDiff)
  {
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    left->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 numSubnodes;
    subnodes->GetLength(&numSubnodes);
    if (numSubnodes != 3 && numSubnodes != 4)
      COMPLAIN(model,
               (100, "Expected <apply/> to have 3 or 4 children(not %u)",
                numSubnodes))
    nsCOMPtr<nsISimpleEnumerator> subnodeEnum;
    subnodes->Enumerate(getter_AddRefs(subnodeEnum));
    nsCOMPtr<nsISupports> applyOpSupports;
    subnodeEnum->GetNext(getter_AddRefs(applyOpSupports));
    nsCOMPtr<nsIDOMNode> applyOp(do_QueryInterface(applyOpSupports));
    applyOp->GetLocalName(leftType);
    // The only type of apply we accept is diff...
    if (!leftType.Equals(NS_LITERAL_STRING("diff")) &&
        !leftType.Equals(NS_LITERAL_STRING("partialdiff")))
    {
      // This may not be a real error, as we have to check both sides.
#ifdef DEBUG_VALIDATE
      printf("<apply> not <diff/> or <partialdiff/> in GetLHSVariableName()\n");
#endif
      return NS_ERROR_FAILURE;
    }
    // We now have two nodes besides the diff operator. One of them is the variable of
    // interest...
    nsCOMPtr<nsISupports> diffArgSupports;
    subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
    nsCOMPtr<nsIDOMElement> diffArg(do_QueryInterface(diffArgSupports));
#ifdef DEBUG_VALIDATE
    printf("Looking for the <diff/> non-bound variable...\n");
#endif
    nsresult ret = NS_ERROR_FAILURE;
    ret = GetLHSVariableName(model, diffArg, leftName, PR_FALSE);
    if (ret == NS_OK)
    {
#ifdef DEBUG_VALIDATE
      printf("Got it.\n");
#endif
    }
    else
    {
#ifdef DEBUG_VALIDATE
      printf("Not the first one, trying the second...\n");
#endif
    }
    subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
    diffArg = do_QueryInterface(diffArgSupports);
    if (ret != NS_OK)
    {
      ret = GetLHSVariableName(model, diffArg, leftName, PR_FALSE);
    }
#ifdef DEBUG_VALIDATE
    if (ret == NS_OK)
      printf("Got it.\n");
#endif
    if (numSubnodes > 3)
    {
      subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
      diffArg = do_QueryInterface(diffArgSupports);
      if (ret != NS_OK)
        ret = GetLHSVariableName(model, diffArg, leftName, PR_FALSE);
#ifdef DEBUG_VALIDATE
      if (ret == NS_OK)
        printf("Got it.\n");
#endif
    }
    if (ret != NS_OK)
      return model->ValidationError(
        NS_LITERAL_STRING("<diff/> with no bound variable")
                                   );
    return NS_OK;
  }
  else if (leftType.Equals(NS_LITERAL_STRING("bvar")) &&
           allowDiff)
  {
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    left->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 numSubnodes;
    subnodes->GetLength(&numSubnodes);
    if (numSubnodes != 1 && numSubnodes != 2)
      COMPLAIN(model, (40, "bvar with %u subnodes, not 1 or 2.", numSubnodes))
    nsCOMPtr<nsISimpleEnumerator> subnodeEnum;
    subnodes->Enumerate(getter_AddRefs(subnodeEnum));
    nsCOMPtr<nsISupports> bvarSupports;

    subnodeEnum->GetNext(getter_AddRefs(bvarSupports));
    nsCOMPtr<nsIDOMNode> bvar(do_QueryInterface(bvarSupports));
    if (GetLHSVariableName(model, bvar, leftName, PR_FALSE) == NS_OK)
      return NS_OK;
    if (numSubnodes > 1)
    {
      subnodeEnum->GetNext(getter_AddRefs(bvarSupports));
      nsCOMPtr<nsIDOMNode> bvar(do_QueryInterface(bvarSupports));
      if (GetLHSVariableName(model, bvar, leftName, PR_FALSE) == NS_OK)
        return NS_OK;
    }
    COMPLAIN(model, (40, "Can't find variable in bvar"));
  }
  else
  {
    // If it is bvar, it is an error, but it is handled & not reported...
    if (leftType.Equals(NS_LITERAL_STRING("bvar")))
      return NS_ERROR_FAILURE;
    // Otherwise, the user should know about this...
    nsCAutoString type;
    CopyUTF16toUTF8(leftType, type);
    return NS_ERROR_FAILURE;
    //COMPLAIN(model,
    //         (100, "Got something other than apply/ci(%s), "
    //          "allowDiff=%u in GetLHSVariableName()",
    //          type.get(), allowDiff))
  }
  return NS_OK;
}

nsresult
mozCellMLVariable::GetBoundVariableName(
                                        mozCellML* mcm,
                                        nsIDOMNode* left,
                                        nsAutoString& leftName
                                       )
{
  nsAutoString leftType;
  left->GetLocalName(leftType);
  if (leftType.Equals(NS_LITERAL_STRING("apply")))
  {
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    left->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 numSubnodes;
    subnodes->GetLength(&numSubnodes);
    if (numSubnodes != 3 && numSubnodes != 4)
    {
      return mcm->ValidationError(NS_LITERAL_STRING("Invalid diff application"));
    }
    nsCOMPtr<nsISimpleEnumerator> subnodeEnum;
    subnodes->Enumerate(getter_AddRefs(subnodeEnum));
    nsCOMPtr<nsISupports> applyOpSupports;
    subnodeEnum->GetNext(getter_AddRefs(applyOpSupports));
    nsCOMPtr<nsIDOMNode> applyOp(do_QueryInterface(applyOpSupports));
    applyOp->GetLocalName(leftType);
    // The only type of apply we accept is diff...
    if (!leftType.Equals(NS_LITERAL_STRING("diff")) &&
        !leftType.Equals(NS_LITERAL_STRING("partialdiff")))
    {
      return mcm->ValidationError(
                  NS_LITERAL_STRING(
           "Bug?: Got non-diff apply but called GetBoundVariableName!!!")
                                 );
    }
    // We now have two nodes besides the diff operator. One of them is the
    // bound variable
    nsCOMPtr<nsISupports> diffArgSupports;
    subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
    nsCOMPtr<nsIDOMNode> diffArg(do_QueryInterface(diffArgSupports));
    nsresult ret = diffArg->GetLocalName(leftName);
    nsresult rv = NS_ERROR_FAILURE;
    if (ret == NS_OK && leftName.Equals(NS_LITERAL_STRING("bvar")))
      rv = GetLHSVariableName(mcm, diffArg, leftName, PR_TRUE);

    subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
    diffArg = do_QueryInterface(diffArgSupports);
    ret = diffArg->GetLocalName(leftName);
    if (ret == NS_OK && leftName.Equals(NS_LITERAL_STRING("bvar")))
    {
      if (rv == NS_OK)
        COMPLAIN(mcm, (40, "Mixed differentials are not supported."))
      else
        rv = GetLHSVariableName(mcm, diffArg, leftName, PR_TRUE);
    }
    if (numSubnodes > 3)
    {
      subnodeEnum->GetNext(getter_AddRefs(diffArgSupports));
      diffArg = do_QueryInterface(diffArgSupports);
      ret = diffArg->GetLocalName(leftName);
      if (ret == NS_OK && leftName.Equals(NS_LITERAL_STRING("bvar")))
      {
        if (rv == NS_OK)
          COMPLAIN(mcm, (40, "Mixed differentials are not supported."))
        else
          return GetLHSVariableName(mcm, diffArg, leftName, PR_TRUE);
      }
    }
    return rv;
  }
  else
  {
    return mcm->ValidationError(
                      NS_LITERAL_STRING(
           "Bug?: GetBoundVariable called on non-apply element!!!"));
  }
  return NS_OK;
}

static bool
VariablesEquivalent(nsInterfaceHashtable<nsStringHashKey,
                    mozCellMLVariable>& allVariables,
                    nsAString& n1, nsAString& n2)
{
  nsCOMPtr<mozCellMLVariable> v1, v2;
  allVariables.Get(n1, getter_AddRefs(v1));
  allVariables.Get(n2, getter_AddRefs(v2));
  return (v1 == v2);
}

nsresult
mozCellMLVariable::DetermineType(
                                 nsInterfaceHashtable<nsStringHashKey,
                                 mozCellMLVariable>& allVariables,
                                 nsIArray* components,
                                 mozCellML* mcm
                                )
{
  // Default to constant...
  type = mozCellMLIVariable::CONSTANT;

  // Get our name...
  nsAutoString vname, vnameN, vnameC;
  GetComponent(vnameC);
  GetName(vnameN);
  vname = vnameC + NS_LITERAL_STRING("::") + vnameN;

  PRUint32 i, l;
  components->GetLength(&l);
  nsCOMPtr<nsISimpleEnumerator> componentsE;
  components->Enumerate(getter_AddRefs(componentsE));
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> componentSupports;
    componentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> componentEl(do_QueryInterface(componentSupports));


    nsAutoString compname;
    componentEl->GetAttribute(NS_LITERAL_STRING("name"), compname);

    nsCOMPtr<nsIDOMNodeList> dnl;
    componentEl->GetChildNodes(getter_AddRefs(dnl));
  
    nsCOMPtr<nsIArray> mathNodes;
    FilterNodeListByNameAndTypeNS(dnl, getter_AddRefs(mathNodes),
                                  nsIDOMNode::ELEMENT_NODE,
                                  MATHML_NS,
                                  NS_LITERAL_STRING("math"));
    nsCOMPtr<nsISimpleEnumerator> mathNodesE;
    mathNodes->Enumerate(getter_AddRefs(mathNodesE));
    PRUint32 j, n;
    mathNodes->GetLength(&n);
    for (
         j = 0;
         j < n;
         j++
        )
    {
      nsCOMPtr<nsISupports> mathElSupports;
      mathNodesE->GetNext(getter_AddRefs(mathElSupports));
      nsCOMPtr<nsIDOMElement> mathEl = do_QueryInterface(mathElSupports);
      nsCOMPtr<nsIDOMNodeList> topLevelAppliesAll;
      mathEl->GetChildNodes(getter_AddRefs(topLevelAppliesAll));
      nsCOMPtr<nsIArray> topLevelApplies;
      FilterNodeListByTypeNS(topLevelAppliesAll,
                             getter_AddRefs(topLevelApplies),
                             nsIDOMNode::ELEMENT_NODE,
                             MATHML_NS);
      PRUint32 numTopApplyNodes;
      topLevelApplies->GetLength(&numTopApplyNodes);
      nsCOMPtr<nsISimpleEnumerator> topLevelAppliesE;
      topLevelApplies->Enumerate(getter_AddRefs(topLevelAppliesE));
      for (PRUint32 k = 0; k < numTopApplyNodes; k++)
      {
        nsCOMPtr<nsISupports> applyNodeSupports;
        topLevelAppliesE->GetNext(getter_AddRefs(applyNodeSupports));
        nsCOMPtr<nsIDOMNode> applyNode(do_QueryInterface(applyNodeSupports));
        PRUint16 topLevelType;
        applyNode->GetNodeType(&topLevelType);
        nsAutoString name;
        applyNode->GetLocalName(name);
        if (!name.Equals(NS_LITERAL_STRING("apply")))
          return mcm->ValidationError(
                                      NS_LITERAL_STRING(
          "Found an element in a math element that is not apply but is ") +
                                      name
                                      );
        nsCOMPtr<nsIDOMNodeList> applyChildrenAll;
        applyNode->GetChildNodes(getter_AddRefs(applyChildrenAll));
        nsCOMPtr<nsIArray> applyChildren;
        FilterNodeListByTypeNS(applyChildrenAll,
                               getter_AddRefs(applyChildren),
                               nsIDOMNode::ELEMENT_NODE,
                               MATHML_NS);
        PRUint32 numApplyChildren;
        applyChildren->GetLength(&numApplyChildren);
        if (numApplyChildren != 3)
          return mcm->ValidationError(
                                      NS_LITERAL_STRING(
          "Found a top-level math apply without 3 child elements.")
                                     );
        nsCOMPtr<nsISimpleEnumerator> applyChildEnum;
        applyChildren->Enumerate(getter_AddRefs(applyChildEnum));
        nsCOMPtr<nsISupports> opNodeSupports;
        applyChildEnum->GetNext(getter_AddRefs(opNodeSupports));
        nsCOMPtr<nsIDOMNode> opNode(do_QueryInterface(opNodeSupports));
        opNode->GetLocalName(name);
        if (!name.Equals(NS_LITERAL_STRING("eq")))
          return mcm->ValidationError(
                                      NS_LITERAL_STRING(
           "Sorry, we only handle top-level math apply nodes which are <eq/>, not "
                                                       ) +
                                      name
                                     );
        // We now have an <apply><eq /><something /><something /></apply>. Classify the
        // "something" variables...
        EqualsArgumentTypes lhsType, rhsType;
        PRUint32 lhsDegree, rhsDegree;
        nsCOMPtr<nsISupports> lhsSupports, rhsSupports;
        applyChildEnum->GetNext(getter_AddRefs(lhsSupports));
        applyChildEnum->GetNext(getter_AddRefs(rhsSupports));
        nsCOMPtr<nsIDOMElement> lhs(do_QueryInterface(lhsSupports)),
                                rhs(do_QueryInterface(rhsSupports));
        nsresult ret = ClassifyEqualsArgument(lhs, &lhsType, &lhsDegree, mcm);
        if (ret != NS_OK)
          return ret;
        ret = ClassifyEqualsArgument(rhs, &rhsType, &rhsDegree, mcm);
        if (ret != NS_OK)
          return ret;
        // We can now use the subclassifications to find the variables to use...
        if (lhsType == APPLY_DIFF && rhsType == APPLY_DIFF)
          return mcm->ValidationError(
                                      NS_LITERAL_STRING(
"Sorry, we don't handle <apply><eq /><apply><diff />...</apply><apply><diff />...</apply></apply> type models."
                                                        )
                                      );
        if (lhsType == APPLY_DIFF || rhsType == APPLY_DIFF)
        {
          nsCOMPtr<nsIDOMElement> tassignTo = (lhsType == APPLY_DIFF) ? lhs : rhs;
          nsCOMPtr<nsIDOMElement> tassignFrom = (lhsType == APPLY_DIFF) ? rhs : lhs;
          PRUint32 tdegree = (lhsType == APPLY_DIFF) ? lhsDegree : rhsDegree;
          nsAutoString leftName;
          ret = GetLHSVariableName(mcm, tassignTo, leftName);
          if (ret != NS_OK)
            return mcm->ValidationError(
                                        NS_LITERAL_STRING(
                   "Assignment has no valid left-hand side."
                                                          )
                                       );
          // After all this, it might not even be the correct variable...
          nsAutoString lcname = compname + NS_LITERAL_STRING("::") + leftName;
#ifdef DEBUG_VARIABLE_CLASSIFY
          {
            nsCAutoString want, got;
            CopyUTF16toUTF8(vname, want);
            CopyUTF16toUTF8(lcname, got);
            printf("Looking for %s, found equation for %s(equiv = %u)\n",
                   want.get(), got.get(),
                   VariablesEquivalent(allVariables, vname, lcname));
          }
#endif
          if (!VariablesEquivalent(
                                   allVariables,
                                   vname,
                                   lcname
                                  ))
          {
            // Check if this is the bound variable...
            nsAutoString boundName;
            ret = GetBoundVariableName(mcm, tassignTo, boundName);
            if (ret != NS_OK)
              return mcm->ValidationError(
                                          NS_LITERAL_STRING(
                         "No valid bound variable found in diff application"
                                                           )
                                         );
            nsAutoString lcname = compname + NS_LITERAL_STRING("::") + boundName;
#ifdef DEBUG_VARIABLE_CLASSIFY
            {
              nsCAutoString want, got;
              CopyUTF16toUTF8(vname, want);
              CopyUTF16toUTF8(lcname, got);
              printf("Looking for %s, found bound %s(equiv = %u)\n",
                     want.get(), got.get(),
                     VariablesEquivalent(allVariables, vname, lcname));
            }
#endif
            if (VariablesEquivalent(
                                    allVariables,
                                    vname,
                                    lcname
                                   ))
            {
              type = mozCellMLIVariable::BOUND;
              assignTo = tassignTo;
              assignFrom = tassignFrom;
              return NS_OK;
            }
            continue;
          }
          degree = tdegree;
          assignTo = tassignTo;
          assignFrom = tassignFrom;
          assignComponent = compname;
          type = mozCellMLIVariable::DIFFERENTIAL;
          return NS_OK;
        }
        else if (lhsType == CI || rhsType == CI)
        {
          // If only one side is a CI, we try that one...
          // XXX for now if both sides are CI, we always assign from the right to
          //     the left. Ideally, we should look at what other options are
          //     available, and use this as a last resort.
          //if (lhsType != rhsType)
          //{
          nsCOMPtr<nsIDOMElement> tassignTo = (lhsType == CI) ? lhs : rhs;
          nsCOMPtr<nsIDOMElement> tassignFrom = (lhsType == CI) ? rhs : lhs;
          nsAutoString leftName;
          GetLHSVariableName(mcm, tassignTo, leftName);
#ifdef DEBUG_VARIABLE_CLASSIFY
          {
            nsCAutoString want, got1, got2;
            CopyUTF16toUTF8(vname, want);
            CopyUTF16toUTF8(leftName, got1);
            nsAutoString rightName;
            GetLHSVariableName(mcm, tassignFrom, rightName);
            CopyUTF16toUTF8(rightName, got2);
            printf("Looking for %s, found simple assignment %s=%s\n",
                   want.get(), got1.get(), got2.get());
          }
#endif
          nsAutoString lcname = compname + NS_LITERAL_STRING("::") + leftName;
          if (!VariablesEquivalent(allVariables, vname, lcname))
            continue;
          assignFrom = tassignFrom;
          assignTo = tassignTo;
          assignComponent = compname;
          type = mozCellMLIVariable::SIMPLE;
          //}
        }
        else
        {
          // We could do Lapack etc... here in some cases, but I don't think its
          // worth it at this stage...
          return mcm->ValidationError(
                       NS_LITERAL_STRING(
             "Sorry, we can't solve models unless there is either a differential variable or a CI on one side."
                                        )
                                     );
        }
      }
    }
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetName(nsAString& name)
{
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("name"), name);
  return ret;
}

NS_IMETHODIMP
mozCellMLVariable::GetComponent(nsAString& name)
{
  nsCOMPtr<nsIDOMNode> componentNode;
  variable->GetParentNode(getter_AddRefs(componentNode));
  nsCOMPtr<nsIDOMElement> component(do_QueryInterface(componentNode));
  nsresult ret = component->GetAttribute(NS_LITERAL_STRING("name"), name);
  return ret;
}

NS_IMETHODIMP
mozCellMLVariable::GetInitialValue(double* initial)
{
  nsAutoString initialStr;
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("initial_value"), initialStr);
  if (ret != NS_OK ||
      initialStr.IsVoid())
    return NS_ERROR_FAILURE;
  nsCString initialCStr;
  CopyUTF16toUTF8(initialStr, initialCStr);
  *initial = PR_strtod(initialCStr.get(), nsnull);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetUnits(nsAString& units)
{
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("units"), units);
  return ret;
}

nsresult
mozCellMLVariable::GetVariableName(nsACString& vn)
{
  vn = varname;
  return NS_OK;
}

nsresult
mozCellMLVariable::SetVariableName(const nsACString& vn)
{
  varname = vn;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::SetInitialValue(double value)
{
  nsPrintfCString ivCString("%e", value);
  nsAutoString ivString;
  CopyUTF8toUTF16(ivCString, ivString);
  return variable->SetAttribute(NS_LITERAL_STRING("initial_value"), ivString);
}

NS_IMETHODIMP
mozCellMLVariable::GetPublicInterface(
                                      nsAString& iface
                                     )
{
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("public_interface"), iface);
  if (ret != NS_OK || iface.IsVoid())
      return NS_ERROR_FAILURE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetPrivateInterface(
                                      nsAString& iface
                                     )
{
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("private_interface"), iface);
  if (ret != NS_OK || iface.IsVoid())
    return NS_ERROR_FAILURE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::IsFlagged(PRUint32 flags, PRBool *isFlagged)
{
  *isFlagged = PR_FALSE;
  if (
      flags &
      (mozICellML::HAS_INITIAL_VALUE|mozICellML::NO_INITIAL_VALUE)
     )
  {
    double initial;
    nsresult res = GetInitialValue(&initial);
    if (res==NS_OK && (flags & mozICellML::NO_INITIAL_VALUE))
      return NS_OK;
    else if (res != NS_OK && (flags & mozICellML::HAS_INITIAL_VALUE))
      return NS_OK;
  }

  if (
      flags &
      (mozICellML::HAS_PUBLIC_INTERFACE|mozICellML::NO_PUBLIC_INTERFACE|
       mozICellML::NO_INTERFACE_IN)
     )
  {
    nsAutoString pubinterface, privinterface;
    nsresult res = GetPublicInterface(pubinterface);
    nsresult res2 = GetPrivateInterface(privinterface);
    if (res==NS_OK && (flags & mozICellML::NO_PUBLIC_INTERFACE))
      return NS_OK;
    else if (res != NS_OK && (flags & mozICellML::HAS_PUBLIC_INTERFACE))
      return NS_OK;
    else if ((flags & mozICellML::NO_INTERFACE_IN) &&
             ((res == NS_OK &&
               pubinterface.Equals(NS_LITERAL_STRING("in"))) ||
              (res2 == NS_OK &&
               privinterface.Equals(NS_LITERAL_STRING("in")))))
      return NS_OK;
  }
  *isFlagged = PR_TRUE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetType(PRUint32* type)
{
  *type = this->type;
  return NS_OK;
}

nsresult
mozCellMLVariable::
GetComputationString(
                     mozCellML* model,
                     nsInterfaceHashtable<nsStringHashKey,mozCellMLVariable>&
                       allVariables,
                     nsACString& code
                    )
{
  if (type != mozCellMLIVariable::DIFFERENTIAL || degree == 1)
    return GetComputationString(
                                model,
                                assignFrom,
                                allVariables,
                                code
                               );
  // This is a higher degree differential, so we need to write a series of
  // unique assignments. Firstly, the main one...
  nsPrintfCString mainCode("x[%u];\n", higher_space_x + degree - 2);
  code += mainCode;
  // Now the code for that uses the expression...
  nsPrintfCString addCode("  dx[%u] = ", higher_space_dx);
  code += addCode;
  GetComputationString(
                       model,
                       assignFrom,
                       allVariables,
                       code
                      );
  code += ";\n";
  PRUint32 i;
  for (i = 3; i < degree; i++)
  {
    nsPrintfCString addCode("  dx[%u] = x[%u];\n",
                            higher_space_dx + i - 2, higher_space_x + i - 3);
    code += addCode;
  }
  return NS_OK;
}

PRUint32
mozCellMLVariable::FetchDegree()
{
  return degree;
}

void
mozCellMLVariable::SetHigherSpace(PRUint32 ixhigher_space, PRUint32 idxhigher_space)
{
  higher_space_x = ixhigher_space;
  higher_space_dx = idxhigher_space;
}

#define IFOP(n) else if (lname.Equals(NS_LITERAL_STRING(#n)))
#define FUNCCONST(n,v) IFOP(n)\
                         code += NS_LITERAL_CSTRING(#v);

#define ROOTCONST(n,v)  else if (name.Equals(NS_LITERAL_STRING(#n)))\
                          code += NS_LITERAL_CSTRING(#v);

#define FUNCOP1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (50, "Apply " #n " should have 2 children not %u.\n", numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "("); \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    }
#define FUNCOP2(n,f) IFOP(n){\
                      if (numSubs != 3)\
                      {\
                        COMPLAIN(model, (50, "Apply " #n " should have 3 children not %u.\n", numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "(");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += ", "; \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                     }
#define FUNCROP1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (50, "Apply " #n " should have 2 children not %u.\n", numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING("1 / " #f "(");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    }
#define FUNCOPR1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (50, "Apply " #n " should have 2 children not %u.\n", numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "(1 / (");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING("))"); \
                    }
#define COPERN(n,o) IFOP(n){\
                    if (numSubs < 2)\
                    {\
                      COMPLAIN(model, (50, "Apply " #n " should have 2+ children not %u.\n", numSubs));\
                      return NS_ERROR_FAILURE;\
                    }\
                    code += NS_LITERAL_CSTRING("(");\
                    e->GetNext(getter_AddRefs(opSupports)); \
                    op = do_QueryInterface(opSupports); \
                    ret = GetComputationString(model, op, allVariables, code); \
                    if (ret != NS_OK) \
                      return ret; \
                    code += NS_LITERAL_CSTRING(")"); \
                    for (PRUint32 x = 2; x < numSubs; x++) \
                    { \
                      code += NS_LITERAL_CSTRING(o "("); \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, code); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    } \
                  }
#define COPER1(n,o) IFOP(n){\
                    if (numSubs != 2)\
                    {\
                      COMPLAIN(model, (50, "Apply " #n " should have 2 children not %u.\n", numSubs));\
                      return NS_ERROR_FAILURE;\
                    }\
                    code += NS_LITERAL_CSTRING("!(");\
                    e->GetNext(getter_AddRefs(opSupports)); \
                    op = do_QueryInterface(opSupports); \
                    ret = GetComputationString(model, op, allVariables, code); \
                    if (ret != NS_OK) \
                      return ret; \
                    code += NS_LITERAL_CSTRING(")"); \
                  }
#define FUNCDEFAULT(n,f,d,m,o) IFOP(n){\
      if (numSubs < 2 || numSubs > 3)\
      {\
        return NS_ERROR_FAILURE;\
      }\
      nsCOMPtr<nsIDOMElement> inside;\
      if (numSubs == 2)\
      {\
        nsCOMPtr<nsISupports> insideSupports; \
        e->GetNext(getter_AddRefs(insideSupports));\
        inside = do_QueryInterface(insideSupports); \
        code += #f "(";\
        ret = GetComputationString(model, inside, allVariables, code);\
        if (ret != NS_OK) \
          return ret; \
        code += ", " #d ")"; \
      }\
      else\
      {\
        nsCOMPtr<nsISupports> p1Supports, p2Supports; \
        e->GetNext(getter_AddRefs(p1Supports)); \
        e->GetNext(getter_AddRefs(p2Supports)); \
        nsCOMPtr<nsIDOMElement> p1(do_QueryInterface(p1Supports)), \
                                p2(do_QueryInterface(p2Supports)), degV; \
        nsAutoString nm1, nm2;\
        p1->GetLocalName(nm1);\
        p2->GetLocalName(nm2);\
        if (nm1.Equals(NS_LITERAL_STRING(#o)))\
        {\
          inside = p2;\
          degV = p1;\
        }\
        else if (nm2.Equals(NS_LITERAL_STRING(#o)))\
        {\
          inside = p1;\
          degV = p2;\
        }\
        else\
        {\
          COMPLAIN(model, (40, "Malformed " #n " invalid child\n"));\
          return NS_ERROR_FAILURE;\
        }\
        nsCOMPtr<nsIDOMNodeList> degSubnodesAll;\
        nsCOMPtr<nsIArray> degSubnodes;\
        degV->GetChildNodes(getter_AddRefs(degSubnodesAll));\
        FilterNodeListByTypeNS(degSubnodesAll, getter_AddRefs(subnodes),\
                               nsIDOMNode::ELEMENT_NODE, MATHML_NS);\
        PRUint32 numDegs;\
        degSubnodes->GetLength(&numDegs);\
        if (numDegs != 1)\
        {\
          COMPLAIN(model, (100, "Malformed " #n " expected 1 " #m " got %u\n", numDegs));\
          return NS_ERROR_FAILURE;\
        }\
        nsCOMPtr<nsISimpleEnumerator> de;\
        degSubnodes->Enumerate(getter_AddRefs(de));\
        nsCOMPtr<nsISupports> ddeSupports; \
        de->GetNext(getter_AddRefs(ddeSupports)); \
        nsCOMPtr<nsIDOMElement> dde(do_QueryInterface(ddeSupports)); \
        code += #f "(";\
        ret = GetComputationString(model, inside, allVariables, code);\
        if (ret != NS_OK) \
          return ret; \
        code += ", " m "(";\
        ret = GetComputationString(model, dde, allVariables, code);\
        if (ret != NS_OK) \
          return ret; \
        code += "))";\
      } \
                             }

nsresult
mozCellMLVariable::
GetComputationString(
                     mozCellML* model,
                     nsIDOMElement* from,
                     nsInterfaceHashtable<nsStringHashKey, mozCellMLVariable>& allVariables,
                     nsACString& code
                    )
{
  nsAutoString name;
  from->GetLocalName(name);
  if (name.Equals(NS_LITERAL_STRING("ci")))
  {
    nsresult ret = GetLHSVariableName(model, from, name, PR_FALSE);
    if (ret != NS_OK)
    {
#ifdef DEBUG_VALIDATE
      printf("GetLHSVariableName on ci failed.\n");
#endif
      return ret;
    }
    nsCOMPtr<mozCellMLVariable> v;
    allVariables.Get(assignComponent + NS_LITERAL_STRING("::") + name,
                     getter_AddRefs(v));
    if (v == nsnull)
    {
      nsCAutoString vn;
      CopyUTF16toUTF8(name + NS_LITERAL_STRING(" in component ") +
                      assignComponent, vn);
      COMPLAIN(model, (100, "Can't find variable %s", vn.get()))
    }
    if (v->type == mozCellMLIVariable::SIMPLE ||
        v->type == mozCellMLIVariable::DIFFERENTIAL)
    {
      nsCAutoString cname;
      v->GetVariableName(cname);
      code += cname;
    }
    else if (v->type == mozCellMLIVariable::CONSTANT)
    {
      double iv = 0;
      v->GetInitialValue(&iv);
      nsPrintfCString cname("%e", iv);
      code += cname;
    }
    else if (v->type == mozCellMLIVariable::BOUND)
      code += NS_LITERAL_CSTRING("y");
    else
      COMPLAIN(model, (40, "Unknown variable type %u\n", v->type))
    return NS_OK;
  }
  else if (name.Equals(NS_LITERAL_STRING("cn")))
  {
    double value;
    GetCNValue(from, &value);
    nsPrintfCString lit("%e", value);
    code += lit;
  }
  else if (name.Equals(NS_LITERAL_STRING("piecewise")))
  {
    nsresult ret;
    // This way of doing things is short but ugly...
    // Note that we assume otherwise is last, otherwise we ignore the rest...
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    from->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE, MATHML_NS);
    PRUint32 numSubs;
    PRBool gotOtherwise = PR_FALSE;
    subnodes->GetLength(&numSubs);
    if (numSubs < 1)
    {
      return model->ValidationError(
                     NS_LITERAL_STRING("Piecewise function with <1 sub.\n"));
    }
    nsCOMPtr<nsISimpleEnumerator> e;
    subnodes->Enumerate(getter_AddRefs(e));
    nsCOMPtr<nsIDOMElement> piece;
    for (PRUint32 n = 0; n < numSubs; n++)
    {
      nsCOMPtr<nsISupports> pieceSupports;
      e->GetNext(getter_AddRefs(pieceSupports));
      piece = do_QueryInterface(pieceSupports);
      nsCOMPtr<nsIDOMNodeList> psubnodesAll;
      nsCOMPtr<nsIArray> psubnodes;
      piece->GetChildNodes(getter_AddRefs(psubnodesAll));
      FilterNodeListByTypeNS(psubnodesAll, getter_AddRefs(psubnodes),
                             nsIDOMNode::ELEMENT_NODE, MATHML_NS);
      PRUint32 numPSubs;
      psubnodes->GetLength(&numPSubs);
      nsCOMPtr<nsISimpleEnumerator> pe;
      psubnodes->Enumerate(getter_AddRefs(pe));
      nsAutoString pieceType;
      piece->GetLocalName(pieceType);
      if (pieceType.Equals(NS_LITERAL_STRING("piece")))
      {
        if (numPSubs != 2)
          COMPLAIN(model, (40, "Piece with %u elements, not 2", numPSubs))
        nsCOMPtr<nsISupports> valueSupports, conditionSupports;
        pe->GetNext(getter_AddRefs(valueSupports));
        pe->GetNext(getter_AddRefs(conditionSupports));
        nsCOMPtr<nsIDOMElement> value(do_QueryInterface(valueSupports)),
                                condition(do_QueryInterface(conditionSupports));
        code += "(";
        ret = GetComputationString(model, condition, allVariables, code);
        if (ret != NS_OK)
          return ret;
        code += ") ? (";
        ret = GetComputationString(model, value, allVariables, code);
        if (ret != NS_OK)
          return ret;
        code += ") : ";
        if (ret != NS_OK)
          return ret;
      }
      else if (pieceType.Equals(NS_LITERAL_STRING("otherwise")))
      {
        gotOtherwise = PR_TRUE;
        if (numPSubs != 1)
          COMPLAIN(model, (40, "<otherwise> with %u elements, not 1", numPSubs))
        nsCOMPtr<nsISupports> valueSupports;
        pe->GetNext(getter_AddRefs(valueSupports));
        nsCOMPtr<nsIDOMElement> value(do_QueryInterface(valueSupports));
        ret = GetComputationString(model, value, allVariables, code);
        break;
      }
    }
    if (!gotOtherwise)
      code += "NAN";
  }
  else if (name.Equals(NS_LITERAL_STRING("apply")))
  {
    nsCOMPtr<nsIDOMNodeList> subnodesAll;
    nsCOMPtr<nsIArray> subnodes;
    from->GetChildNodes(getter_AddRefs(subnodesAll));
    FilterNodeListByTypeNS(subnodesAll, getter_AddRefs(subnodes),
                           nsIDOMNode::ELEMENT_NODE, MATHML_NS);
    PRUint32 numSubs;
    subnodes->GetLength(&numSubs);
    if (numSubs < 1)
      return model->ValidationError(NS_LITERAL_STRING("<apply> with no child elements."));
    nsCOMPtr<nsISimpleEnumerator> e;
    subnodes->Enumerate(getter_AddRefs(e));
    nsCOMPtr<nsISupports> opSupports;
    e->GetNext(getter_AddRefs(opSupports));
    nsCOMPtr<nsIDOMElement> op(do_QueryInterface(opSupports));
    nsAutoString lname;
    op->GetLocalName(lname);
    nsresult ret;
    // We now have to look at the operator, and convert to a C operator...
    if (lname.Equals(NS_LITERAL_STRING("quotient")))
    {
      if (numSubs != 3)
        COMPLAIN(model, (50, "quotient with %u != 3 arguments.", numSubs))
      code += "floor(";
      e->GetNext(getter_AddRefs(opSupports));
      op = do_QueryInterface(opSupports);
      ret = GetComputationString(model, op, allVariables, code);
      if (ret != NS_OK)
        return ret;
      code += "/";
      e->GetNext(getter_AddRefs(opSupports));
      op = do_QueryInterface(opSupports);
      ret = GetComputationString(model, op, allVariables, code);
      if (ret != NS_OK)
        return ret;
      code += ")";
    }
    FUNCCONST(pi, M_PI)
    FUNCCONST(exponentiale, M_E)
    FUNCCONST(eulergamma, 0.5772156649)
    FUNCCONST(notanumber, NAN)
    FUNCCONST(true, 1.0)
    FUNCCONST(false, 0.0)
    COPERN(divide, "/")
    COPERN(minus, "-")
    COPERN(plus, "+")
    COPERN(times, "*")
    COPERN(and, "&&")
    COPERN(or, "||")
    COPER1(not, "!")
    COPERN(eq, "==")
    COPERN(neq, "!=")
    COPERN(gt, ">")
    COPERN(lt, "<")
    COPERN(geq, ">=")
    COPERN(leq, "<=")
    FUNCOP2(xor, logical_xor)
    FUNCOP2(implies, logical_implies)
    FUNCOP1(factorial, factorial)
    FUNCOP1(exp, exp)
    FUNCOP2(power, pow)
    FUNCDEFAULT(root, pow, 0.5, "1/", degree)
    FUNCOP1(floor, floor)
    FUNCOP1(ceiling, ceil)
    FUNCOP1(abs, abs)
    FUNCOP1(ln, log)
    FUNCDEFAULT(log, arbitrary_log, 10, "", logbase)
    FUNCOP2(rem, get_remainder)
    FUNCOP1(sin, sin)
    FUNCOP1(cos, cos)
    FUNCOP1(tan, tan)
    FUNCROP1(csc, sin)
    FUNCROP1(sec, cos)
    FUNCROP1(cot, tan)
    FUNCOP1(sinh, sinh)
    FUNCOP1(cosh, cosh)
    FUNCOP1(tanh, tanh)
    FUNCROP1(csch, sinh)
    FUNCROP1(sech, cosh)
    FUNCROP1(coth, tanh)
    FUNCOP1(arcsin, arcsin)
    FUNCOP1(arccos, arccos)
    FUNCOP1(arctan, arctan)
    FUNCOP1(arccosh, arccosh)
    FUNCOPR1(arccot, arctan)
    FUNCOPR1(arccoth, arctanh)
    FUNCOPR1(arccsc, arcsin)
    FUNCOPR1(arccsch, arcsinh)
    FUNCOPR1(arcsec, arccos)
    FUNCOPR1(arcsech, arccosh)
    FUNCOP1(arcsinh, arcsinh)
    FUNCOP1(arctanh, arctanh)
    else
    {
      nsCAutoString cname;
      CopyUTF16toUTF8(lname, cname);
      COMPLAIN(model, (40, "Unknown operator in MathML: %s", cname.get()))
    }
    return NS_OK;
  }
  ROOTCONST(pi, M_PI)
  ROOTCONST(exponentiale, M_E)
  ROOTCONST(eulergamma, 0.5772156649)
  ROOTCONST(notanumber, NAN)
  ROOTCONST(true, 1.0)
  ROOTCONST(false, 0.0)
  else
  {
    nsCAutoString fv;
    CopyUTF16toUTF8(name, fv);
    COMPLAIN(model,
             (100, "MathML element isn't ci, cn, apply, or piecewise(%s)", fv.get())
            )
    return NS_ERROR_FAILURE;
  }
  return NS_OK;
}

NS_IMPL_ISUPPORTS0(mozCellMLVariablePromise)

mozCellMLVariablePromise::mozCellMLVariablePromise(
                                                   nsAString& party1,
                                                   nsAString& party2
                                                  )
{
  parties.AppendString(party1);
  parties.AppendString(party2);
}

void
mozCellMLVariablePromise::AddParty(nsAString& party)
{
  if (redundantBy != nsnull)
    redundantBy->AddParty(party);
  parties.AppendString(party);
}

nsStringArray&
mozCellMLVariablePromise::SetRedundant(mozCellMLVariablePromise* r)
{
  nsStringArray& ret = *(nsStringArray*)nsnull;
  if (redundantBy != nsnull)
    ret = redundantBy->SetRedundant(r);
  else
    ret = parties;
  redundantBy = r;
  return ret;
}

void
mozCellMLVariablePromise::MergeIn(mozCellMLVariablePromise* merge)
{
  if (redundantBy != nsnull)
    redundantBy->MergeIn(merge);
  else
  {
    nsStringArray& mergeArray = merge->SetRedundant(this);
    PRUint32 i, l;
    nsAutoString party;
    for (i = 0, l = mergeArray.Count(); i < l; i++)
    {
      mergeArray.StringAt(i, party);
      parties.AppendString(party);
    }
  }
}

nsresult
mozCellMLVariablePromise::Resolve(
                                  mozCellML* model,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       mozCellMLVariable>& table
                                 )
{
  if (redundantBy != nsnull)
    return NS_OK;
  nsAutoString party;
  nsCOMPtr<mozCellMLVariable> var;
  PRUint32 i, l;
  for (i = 0, l = parties.Count(); i < l; i++)
  {
    parties.StringAt(i, party);

    table.Get(party, getter_AddRefs(var));
    if (var == nsnull)
      continue;
    for (i = 0, l = parties.Count(); i < l; i++)
    {
      parties.StringAt(i, party);
      table.Put(party, var);
    }
    return NS_OK;
  }
  nsAutoString party1, party2;
  parties.StringAt(0, party1);
  parties.StringAt(1, party2);

  nsCAutoString cparty1, cparty2;
  CopyUTF16toUTF8(party1, cparty1);
  CopyUTF16toUTF8(party2, cparty2);
  COMPLAIN(model,
           (100, "Unresolved connection involving %s and %s",
            cparty1.get(), cparty2.get()));
}
